// TODO: populate DOM with HTML strings for speed (less readable, though)

var controller = function(controller, jQuery) {
////// visible ///////////////////////////////////////////////////////////
    var visible = {};


    // animation speed ///////////////////////////////////////////////////
    visible.animationSpeed = 0;
    visible.animationDuration = 0;
    visible.animationDelay = 0;
    visible.toggleOnDelay = 0;
    visible.toggleOffDelay = 0;
    
    
    // state /////////////////////////////////////////////////////////////
    visible.editID = 0;
    
    
    // methods ///////////////////////////////////////////////////////////
    visible.init = function() {
        // fetch data to display
        viewData = model.viewData(true, true);

        // prepare interface
        prepareDOM();
        prepareTransitions();
        prepareHandlers();
        resetLinkActionFlags();
        
        // put everything in place
        redraw(true, true);
        
        // TODO refactor - initialize number of times "decided" to 0
        utils.setStorageItem("decided", 0)
        
    };
    
    
    visible.toggleItem = function($item, delay, show) {
        if (show) {
            if (model.hidden[$item.attr("id")] === undefined) {
                $item.show();
                
                if (!displayDetails) {
                    $item.children(".detail").fadeOut();
                    $item.height($.browser.mozilla ? "1.75em" : "1.25em");
                }
                
                setTimeout(function() {
                    if($item.hasClass("shadow"))
		            	$item.css("opacity", "0.5");
		        	else
		            	$item.css("opacity", "1"); // TODO verify

                    // some item was hidden or shown, update review button
                    updateReviewButton();
                }, delay);
            }
        } else {
            setTimeout(function() {
                $item.css("opacity", 0);
                
                setTimeout(function() {
                    $item.hide();
                    $item.removeClass("undecided-hover item-hover");
                    // TODO this will not appear real-time because identical items are not immediately hidden when merged
                    // some item was hidden or shown, update review button
                    updateReviewButton();
                }, delay);
            }, delay);
        }
    };
    
    
////// hidden ////////////////////////////////////////////////////////////
    
    // pseudo constants //////////////////////////////////////////////////
    var LIST_1_UNIQUE_INDEX = 0;
    var LIST_1_INDEX = 1;
    var LIST_IDENTICAL_INDEX = 2;
    var LIST_2_INDEX = 3;
    var LIST_2_UNIQUE_INDEX = 4;
    var NUM_ROWS = {};
    var NUM_COLUMNS = 5;

    var STATE_SEPARATE = 0;
    var STATE_IDENTICAL = 1;
    var STATE_UNIQUE = 2;
    var STATE_SIMILAR = 3;
    var STATE_COMPACT = 4;
    var NUM_STATES = 5;
    
    var BASE_ANIMATION_DURATION = 800;
    var PANEL_DELAY = BASE_ANIMATION_DURATION * 2;
    var ANIMATION_SPEED_0 = "__ANIMATION_SPEED_0__";
    var ANIMATION_SPEED_1 = "__ANIMATION_SPEED_1__";
    var ANIMATION_SPEED_2 = "__ANIMATION_SPEED_2__";
    var ANIMATION_SPEED_3 = "__ANIMATION_SPEED_3__";
    var ANIMATION_SPEED_4 = "__ANIMATION_SPEED_4__";
    var ANIMATION_SPEED_5 = "__ANIMATION_SPEED_5__";
    var ANIMATION_SPEED_COEFFICIENTS = [0, 1.5, 1.25, 1, 0.75, 0.5];
    var CUSTOM_EASE_IN = "cubic-bezier(0.4, 0.2, 0.1, 0.9)";
    
    var VERSION_FULL = "__VERSION_FULL__";
    var VERSION_BASELINE = "__VERSION_BASELINE__";
    var VERSION_LINK_ONLY = "__VERSION_LINK_ONLY__";
    
    var AUTO_ANIMATE_ON = "__AUTO_ANIMATE_ON__";
    var AUTO_ANIMATE_OFF = "__AUTO_ANIMATE_OFF__";
    var AUTO_ANIMATE_END = "__AUTO_ANIMATE_END__";

    var WHITE = "#fafafa";
    var NEUTRAL_GRAY = "#f0f0f0";
    var CONTRAST_GRAY = "#d9d6d0";
    
    // internal state ////////////////////////////////////////////////////
    var viewData = [];
    var hoverSet = {};

    var groups = [];

    var state = STATE_SEPARATE;
    
    // stores item locations for different states
    // 
    var positions = {};

    var keyframes = {};
    
    var linkAction = {};
    
    var displayDetails = true;
    var autoAnimate = AUTO_ANIMATE_ON;
    var version = VERSION_FULL;
    
    // initialization ////////////////////////////////////////////////////
    function resetState() {
        state = STATE_SEPARATE;
        viewData = [];
        hoverSet = {};
    
        groups = [];
    
        positions = {};
        keyframes = {};
        
        linkAction = {};
    }
    
    function changeDataset(dataset) {
        model.init(dataset);
        resetState();
        visible.init();

        logger.log(logger.EVENT_DATASET_CHANGE, dataset);
    }
    
    function prepareDOM() {
        // clear
        $(".backdrop thead").remove();
        $(".backdrop tbody").remove();
        $("#app .item").remove();
        
        // create actual objects
        populateBackdrop($(".backdrop"));
        populateItems($(".items"));
        
        // TODO: faster / neater to wrap all shadows in a separate div?
        
        // start shadowing
        $(".shadow").hide();
        
        // hide unused column headers; avoid unexpected behavior
        $(".backdrop th:nth-child(odd) .col-header").hide();
        $(".conditional").hide();
        
        // indicate current config
        $("select[name='dataset']").val(model.dataset);
        $("#app .annotation").text("[ case : " +
                model.dataset.toLowerCase().replace(/_/g, " ")
                .replace(/dataset/, "").trim() +
                " ]");
        $("select[name='version']").val(version);
        $("select[name='autoAnimate']").val(autoAnimate);
        $("select[name='speed']").val(visible.animationSpeed);
        
        // initialize the number of drugs to act on
        updateReviewButton();
        
    }
    
    function prepareTransitions() {
        // calculate column positions exactly once; they won't change
        calculateColumns();

        // row positions, however, will
        calculatePositions();
        
        // prepare animation speed
        setAnimationSpeed(ANIMATION_SPEED_3);
    }
    
    function prepareHandlers() {
        // clear
        $("*").not(".backdrop *").off();
        
        // panel navigation
        $(".forward").click(function() {
        	
        	// reset dataset when moving forward (remove added/edited items)
        	model.init($("select[name='dataset']").val());
	        resetState();
	        visible.init();
	        
            if ($(this).is("#intro .forward")) {
            	// if the demo is in the initial state
                if (state === STATE_SEPARATE &&
                        model.accepted.length === 0 &&
                        model.rejected.length === 0) {
                    logger.log(logger.EVENT_DEMO_START, "dataset:" + model.getDatasetShortName(model.dataset)
                    								+ ",version:" + getVersionShortName(version));
                } else {
                    logger.log(logger.EVENT_DEMO_RESUME, "");
                }
            }
            panels.showCenter();
            
            // reset group by based on version
            if(version === VERSION_LINK_ONLY) {
            	updateGroupBy("__ATTR_DRUG_CLASS__");
            } else {
            	updateGroupBy("");
            }
            
            
            if (autoAnimate === AUTO_ANIMATE_ON) {
                setTimeout(function() {
                    if (version === VERSION_FULL) {
                        changeState(STATE_COMPACT);
                    }
                }, PANEL_DELAY);
            } else if (autoAnimate === AUTO_ANIMATE_END) {
            	setTimeout(function() {
                    if (version === VERSION_FULL) {
                    	setAnimationSpeed(ANIMATION_SPEED_0);
                        changeState(STATE_COMPACT);
                    }
                }, 0);
            }
        });
        
        $(".backward").click(function() {
        	logger.log( logger.EVENT_DEMO_RESUME, "");
            panels.showCenter();
        });
        
        // item action
        $("#app .item").mousedown(mousedownHandler)
                .mouseover(mouseoverHandler)
                .mouseout(mouseoutHandler)
                .bind("contextmenu", function() {
                    // free right click for item action
                    return false;
                });
        
        // basic controls
        $(".compare").click(function() {
            if (version === VERSION_FULL) {
                logger.log(logger.EVENT_CLICKED, "compare"); // TODO use button constants?
            }
            changeState(STATE_COMPACT);
        });

        $(".review").click(function() {
            logger.log(logger.EVENT_CLICKED, "sign off");

            // save decisions
            var accepted = [];
            var rejected = [];
            var undecided = [];
            var acceptedIds = [];
            
            acceptedIds.push(state);
            for (var i = 0; i < model.accepted.length; i++) {
                accepted.push(logger.itemString(model.accepted[i]));
            	acceptedIds.push(model.accepted[i]);
            }
            
            for (var i = 0; i < model.rejected.length; i++) {
                rejected.push(logger.itemString(model.rejected[i]));
            }
            
            for (var i = 0; i < model.undecided.length; i++) {
                undecided.push(logger.itemString(model.undecided[i]));
            }
            
            // dump current item decisions (accepted, rejected, undecided)
            logger.log(logger.EVENT_LIST_ACCEPTED, accepted);
            logger.log(logger.EVENT_LIST_REJECTED, rejected);
            logger.log(logger.EVENT_LIST_UNDECIDED, undecided);
            logger.log(logger.DATA_DEMO_END_STATE, acceptedIds);

			// if still items to review - show popup
			if (calculateNumRemaining() > 0) {
				//TODO make 2s popup
				alert("Still " + calculateNumRemaining() + 
					" drugs to act upon before signing off"); 
			} else {
				logger.log(logger.EVENT_SIGNED_OFF, "");
				// return to initial panel
            	setTimeout(function() {
            		// prepare patient summary
		            saveItemActions();
		            window.open("summary.html");
        			
        			resetDecisions();
	                updateReviewButton();
	                panels.showLeft();
	            }, PANEL_DELAY / 2); // TODO consider delay time
	            // TODO consider feedback saying the lists were saved
			}
			
		}); // end of $(".review").click

        $(".retry").click(function() {
        	logger.log(logger.EVENT_CLICKED, "reset");
            resetDecisions();
            redraw(true, true);
        });
        
        // options panel
        $(".options").click(function() {
        	logger.log(logger.EVENT_CLICKED, "options");
        	
            if ($(this).hasClass("active")) {
                $(this).text("show options").removeClass("active");

                $(".options-panel").css("transition",
                        "transform 0.3s, opacity 0.2s")
                        .removeClass("show");
                $(".options-panel").css("-webkit-transition",
                        "-webkit-transform 0.3s, opacity 0.2s")
                        .removeClass("show");
                $(".options-panel").css("-moz-transition",
                        "-moz-transform 0.3s, opacity 0.2s")
                        .removeClass("show");
                $(".options-panel").css("-ms-transition",
                        "-ms-transform 0.3s, opacity 0.2s")
                        .removeClass("show");
                $(".options-panel").css("-o-transition",
                        "-o-transform 0.3s, opacity 0.2s")
                        .removeClass("show");
                 
                $("#app > .content").css("transition", "height 0.3s");
                $("#app > .content").css("-webkit-transition",
                        "height 0.3s");
                $("#app > .content").css("-moz-transition",
                        "height 0.3s");
                $("#app > .content").css("-ms-transition",
                        "height 0.3s");
                $("#app > .content").css("-o-transition", "height 0.3s");

                $(".add, .edit").css("transition",
                        "top 0.3s, width 0.2s");
                $(".add, .edit").css("-webkit-transition",
                        "top 0.3s, width 0.2s");
                $(".add, .edit").css("-moz-transition",
                        "top 0.3s, width 0.2s");
                $(".add, .edit").css("-ms-transition",
                        "top 0.3s, width 0.2s");
                $(".add, .edit").css("-o-transition",
                        "top 0.3s, width 0.2s");
                

                redraw(true, true, true);
            } else {
                $(this).text("hide options").addClass("active");
                
                $(".options-panel").css("transition",
                        "transform 0.2s, opacity 0.2s")
                        .addClass("show");
                $(".options-panel").css("-webkit-transition",
                        "-webkit-transform 0.2s, opacity 0.2s")
                        .addClass("show");
                $(".options-panel").css("-moz-transition",
                        "-moz-transform 0.2s, opacity 0.2s")
                        .addClass("show");
                $(".options-panel").css("-ms-transition",
                        "-ms-transform 0.2s, opacity 0.2s")
                        .addClass("show");
                $(".options-panel").css("-o-transition",
                        "-o-transform 0.2s, opacity 0.2s")
                        .addClass("show");
                         
                $("#app > .content").css("transition", "height 0.2s");
                $("#app > .content").css("-webkit-transition",
                        "height 0.2s");
                $("#app > .content").css("-moz-transition",
                        "height 0.2s");
                $("#app > .content").css("-ms-transition",
                        "height 0.2s");
                $("#app > .content").css("-o-transition", "height 0.2s");

                $(".add, .edit").css("transition",
                        "top 0.2s, width 0.2s");
                $(".add, .edit").css("-webkit-transition",
                        "top 0.2s, width 0.2s");
                $(".add, .edit").css("-moz-transition",
                        "top 0.2s, width 0.2s");
                $(".add, .edit").css("-ms-transition",
                        "top 0.2s, width 0.2s");
                $(".add, .edit").css("-o-transition",
                        "top 0.2s, width 0.2s");
                
                redraw(true, true, true);
            }
        });
        
        // config
        $("select[name='dataset']").change(function() {
            if ($(this).val() !== model.dataset) {
                changeDataset($(this).val());
            }
        });
        
        $("select[name='version']").change(function() {
            version = $(this).val();
            
            if ($(this).val() === VERSION_BASELINE) {
                state = STATE_SEPARATE;
                changeState(STATE_SEPARATE);
            } else {
                adjustAnimationControls(state);
            }
            
            resetLinkActionFlags();
            logger.log(logger.EVENT_VERSION_CHANGE, version);
        });
        
        $("select[name='autoAnimate']").change(function() {
            autoAnimate = $(this).val();
        });
        

        // when clicked, cycles the current groupby layout used
        $(".cycle_groupby").click(function () {
        	logger.log(logger.EVENT_CLICKED, "cycle_groupby");
        	
        	updateGroupBy($(this).attr("value"));
        })
        
        // help modal
        $(".help-modal").css("top", $("header").height());

        $(".help").click(function() {
        	logger.log(logger.EVENT_CLICKED, "help");
            $(".help-modal").addClass("show");
        });

        $(".help-close").click(function() {
        	logger.log(logger.EVENT_CLICKED, "help_close");
            $(".help-modal .welcome").remove();
            $(".help-modal").removeClass("show");
        });
        
        // alert modal
        $("#alert-close").click(function() {
        	logger.log(logger.EVENT_CLICKED, "alert_close");
            $(".alert-modal").removeClass("show");
        });
        
        // options
        $("select[name='groupBy']").change(function() {
        	updateGroupBy($(this).attr("value"));
        	
        	/*
            model.groupBy = $(this).val();
            
            // force; grouping in compact looks strange
            if (state > STATE_SIMILAR) {
                state = STATE_SIMILAR;
            }
            redraw(true, true);
            */
        });
        
        $("input[name='multigroup']").click(function() {
        	logger.log(logger.EVENT_CLICKED, "multigroup");
        	
            model.multigroup = this.checked;
            
            // automatically change to grouping by drug class
            // update the grouping in the options (using value)
			$("select[name='groupBy']").val("__ATTR_DRUG_CLASS__");
			
			// trigger layout change - redraw(true, true) handled in here too
			$("select[name='groupBy']").change();
        });
        
        $("select[name='speed']").change(function() {
            setAnimationSpeed($(this).val());
        });

        $("input[name='filterOn']").keyup(function() {
            model.filterOn = $(this).val();
            redraw(true, true);
        });

        $(".separate").click(function() {
        	logger.log(logger.EVENT_CLICKED, "separate");
            changeState(STATE_SEPARATE);
        });

        $(".identical").click(function() {
        	logger.log(logger.EVENT_CLICKED, "identical");
            changeState(STATE_IDENTICAL);
        });

        $(".unique").click(function() {
        	logger.log(logger.EVENT_CLICKED, "unique");
            changeState(STATE_UNIQUE);
        });

        $(".similar").click(function() {
        	logger.log(logger.EVENT_CLICKED, "similar");
            changeState(STATE_SIMILAR);
        });

        $(".compact").click(function() {
        	logger.log(logger.EVENT_CLICKED, "compact");
            changeState(STATE_COMPACT);
        });

        $(".grayout").click(function() {
        	logger.log(logger.EVENT_CLICKED, "grayout");
            if (!$(this).hasClass("active")) {
                $(".remove").removeClass("active");
                $(this).addClass("active");

                model.afterAction = model.AFTER_ACTION_GRAYOUT;
                redraw(true, true);
            }
        });

        $(".remove").click(function() {
        	logger.log(logger.EVENT_CLICKED, "remove");
            if (!$(this).hasClass("active")) {
                $(".grayout").removeClass("active");
                $(this).addClass("active");

                model.afterAction = model.AFTER_ACTION_REMOVE;
                redraw(true, true);
            }
        });
        
        $("select[name='displayName']").change(function() {
            model.displayName = $(this).val();
            
            $("#app .item").each(function() {
                var id = this.id;
                
                $("#" + id + " .name").fadeOut("fast")
                        .text(model.items[id].name)
                        .fadeIn("fast");
            });
            redraw(true, true);
        });
        
        $("input[name='displayDetails']").click(function() {
        	logger.log(logger.EVENT_CLICKED, "displayDetails");
            displayDetails = this.checked;
            redraw(true, true);
        });
        
        // main panel: item modification
        $(".add").click(function() {
        	logger.log(logger.EVENT_CLICKED, "add");
            resetModifyPanel("add")
            panels.showRight("modify");
        });
        
        $(".edit").click(function() {
        	logger.log(logger.EVENT_CLICKED, "edit");
            resetModifyPanel("edit");
            panels.showRight("modify");
        });
        
        // preview
        $("input, textarea").keyup(function() {
            updatePreview($(this));
        });
        
        // modify panel
        $("#modify .save").click(function() {
        	logger.log(logger.EVENT_CLICKED, "save_modified");
            saveModifiedItem($("#modify h1").text());
            // $("#modify .response").addClass("show");
            $(".back").click()
        });
        
        // modify / review backwards shortcut
        $(".back").click(function() {
        	logger.log( logger.EVENT_DEMO_RESUME, "");
            panels.showCenter();
        });
        
        // exit button
        $(".cancel").click(function() {
        	resetDecisions();
            updateReviewButton();
            panels.showLeft();
        });
        
        
		// import dialog
		$("#import-dialog").dialog({
			autoOpen : false,
			height : 300,
			width : 350,
			modal : true,
			buttons : {
				"Submit" : function() {
					var text = $("#import-data").val();
					
					var acceptedIds = text.split(",");
					
					if(acceptedIds.length < 2) {
						alert("information missing (need dataset, state, and ids)");
					} else {
						var dataset = model.getDatasetLongName(acceptedIds[0]);
						var savedState = acceptedIds[1];
						acceptedIds.splice(0, 2); // remove first 2 elements
						
						if(!dataset) {
							alert("invalid dataset");
						} else {
							// reset interface / initialize parameters
							version = VERSION_FULL;
							model.init(dataset);
					        resetState();
					        visible.init();
					        updateGroupBy("");
					        
					        // jump to saved state
					        setAnimationSpeed(ANIMATION_SPEED_0);
					        transition(savedState, savedState);
							
							// apply saved decisions
							for (var id in model.items) {
								var $item = $("#" + id);
								
								if($.inArray(id, acceptedIds) >= 0) {
									actOnItem(id, "accepted");
								} else {
									actOnItem(id, "rejected");
								}
							}
							
				            panels.showCenter();
						}
					}
					
					$(this).dialog("close");
				},
				Cancel : function() {
					$(this).dialog("close");
				}
			}
		}); 

        
        // hidden controls (shortcuts)
        $(window).unbind("keydown")		// for some reason, multiple keydowns firing, remove previous ones
        .keydown(function(event) {
            
            switch (event.which) {
                case 48:                            // the '0' key
                    setAnimationSpeed(ANIMATION_SPEED_0);
                    break;
                case 49:                            // the '1' key
                    setAnimationSpeed(ANIMATION_SPEED_1);
                    break;
                case 50:                            // the '2' key
                    setAnimationSpeed(ANIMATION_SPEED_2);
                    break;
                case 51:                            // the '3' key
                    setAnimationSpeed(ANIMATION_SPEED_3);
                    break;
                case 52:                            // the '4' key
                    setAnimationSpeed(ANIMATION_SPEED_4);
                    break;
                case 53:                            // the '5' key
                    setAnimationSpeed(ANIMATION_SPEED_5);
                    break;
                case 76:                            // the 'l' key
                    logger.dump();
                    window.open("log_summary.html");
                    break;
                case 77:							// the 'm' key
                	// toggle multigroup
                	
                	if($("#chk_mg").attr("checked")) {
			            $("#chk_mg").attr('checked', false);
			        } else {
			            $("#chk_mg").attr('checked', true);
			        }
                	
                	// don't trigger .click() because it automatically toggles, also this avoids the registering a click
                	model.multigroup = $("#chk_mg").attr("checked");
		            // automatically change to grouping by drug class
		            // update the grouping in the options (using value)
					$("select[name='groupBy']").val("__ATTR_DRUG_CLASS__");
					
					// trigger layout change - redraw(true, true) handled in here too
					$("select[name='groupBy']").change();
			
                	break;
            	case 78:							// the 'n' key
            		// group by none
            		updateGroupBy("");
            		break;
        		case 68:							// the 'd' key
        			// group by drug class
        			updateGroupBy("__ATTR_DRUG_CLASS__");
            		break;
        		case 82:							// the 'r' key
        			// group by route
        			updateGroupBy("__ATTR_ROUTE__");
            		break;
            }
        });
        
        $(window).unbind("keyup")		// for some reason, multiple keydowns firing, remove previous ones
        .keyup(function(event) {
        	switch (event.which) {
        		case 73:							// the 'i' key
        			// open the import dialog - for experimenter review
        			openImportDialog();
        			break;
			}
        });
        
		// count scrolls
		var lastScrollTop = 0;
		$(".content").unbind("scroll")
		.bind('scroll', function() {
			var st = $(this).scrollTop();
			if (st > lastScrollTop){
				// downscroll code
   				logger.log(logger.EVENT_SCROLLED, "down");
   			} else {
  				// upscroll code
  				logger.log(logger.EVENT_SCROLLED, "up");
   			}
   			lastScrollTop = st;
		});
        
        // resize with window
        $(window).resize(function() {
            setTimeout(function() {
                redraw(true, true, true);
            }, visible.animationDuration + visible.animationDelay);
        });
    }
    
    function populateBackdrop($container) {
        var $thead = $("<thead></thead>");
        $thead.attr("class", "bg-accent");
        
        var $tr = $("<tr></tr>");
        
        for (var i = 0; i < NUM_COLUMNS; i++) {
            var $th = $("<th></th>");
            var $header = $("<div></div>");
            var $name = $("<div></div>");
            
            // column name
            switch (i) {
                case LIST_1_UNIQUE_INDEX:
                    $name.text(model.list1.name + " unique");
                    break;
                case LIST_1_INDEX:
                    $name.text(model.list1.name)
                            .append("<span class='conditional'> similar" +
                                    "</span>");
                    break;
                case LIST_IDENTICAL_INDEX:
                    $name.text("Identical");
                    break;
                case LIST_2_INDEX:
                    $name.text(model.list2.name)
                            .append("<span class='conditional'> similar" +
                                    "</span>");
                    break;
                case LIST_2_UNIQUE_INDEX:
                    $name.text(model.list2.name + " unique");
                    break;
            }
            $name.attr("class", "name");
            $header.attr("class", "col-header");
            $header.append($name);
            
            // if 2 col versions, move 1 and 3 closer
            if(version != VERSION_FULL) {
        		if(i == LIST_1_INDEX){
        			$th.css("left", "10%");
        			$th.css("position", "relative");
        		} else if(i == LIST_2_INDEX) {
        			$th.css("right", "10%");
        			$th.css("position", "relative");
        		}
            } 
            
            // column actions
            var $action = $("<div></div>");
            var $ul = $("<ul></ul>");
            
            /* 
             * Note: the key names here (e.g "keep", "reject", "clear") 
             * become the text for the displayed buttons (for convenience).
             * Furthermore, they receive a class with the same name.
             * So, changes in button text require changes in style.css
             */
            var actions = {
                keep : "accepted", 
                reject : "rejected",
                clear : "undecided"
            };
            
            for (var action in actions) {
                var $li = $("<li></li>");
                var $a = $("<a></a>");
                
                $a.attr("class", action);
                if(action == "clear") {
                	$a.text(action);
                } else {
                	$a.text(action + " rest");	
                }                
                $a.click({ index : i, dst : actions[action] },
                        processColumn);
                $li.append($a);
                $ul.append($li);
            }
            $action.attr("class", "action");
            $action.append($ul);
            $header.append($action);
            $th.append($header);
            $tr.append($th);
        }
        $thead.append($tr);
        $container.append($thead);
        $container.append("<tbody></tbody>");
        
    }
    
    function populateItems($container) {
        
        // TODO: populate shadows first for less awkward overlaps
        
        // populate all items and shadows
        for (var id in model.items) {
            var item = model.items[id];
            var differences = [];
            
            var $item = $("<div></div>");
            $item.attr("id", id);
            $item.attr("class", "item undecided");
            
            if (item.isShadow) {
                $item.addClass("shadow");
            }
            
            /*
             * Highlighting a block element name allows for text-overflow,
             * but typically results in unnecessary color due to
             * 100% width. Thus wrap an inline name in a block element to
             * enable text-overflow and more natural name highlighting.
             */
            var $header = $("<div></div>");
            var $name = $("<span></span>");
            
            $header.attr("class", "header");
            $name.attr("class", "name");
            $name.text(item.name);
            
            if (id in model.similar) {
                differences = model.similar[id].differences;
                
                if (differences.indexOf(model.ATTR_NAME) !== -1) {
                    $name.addClass("difference");
                }
            }
            var $detail = $("<div></div>");
            $detail.attr("class", "detail");
            
            for (var attribute in item.attributes) {
                if (model.attributes[attribute].display) {
                    var $attribute = $("<span></span>");
                    $attribute.text(item.attributes[attribute]
                            .toString());
                    
                    if (id in model.similar) {
                        if (differences.indexOf(attribute) !== -1) {
                            $attribute.addClass("difference");
                        }
                    }
                    $detail.append($attribute);
                }
            }
            $header.append($name);
            $item.append($header);
            $item.append($detail);
            $container.append($item);
        }
        
    }
    
    
    // position calculation //////////////////////////////////////////////
    function calculateColumns() {
        for (var id in model.items) {
            var checkID = id;
            var checkItem = model.items[id];
            var side = (checkItem.listID === model.list1.id) ? -1 : 1;
            var threshold = STATE_UNIQUE;
            var initialCol = Math.floor(NUM_COLUMNS / 2) + side;
            var finalCol = initialCol + side;
            
            if (checkItem.isShadow) {
                // shadow column based on original
                checkID = model.getShadowed(id);
            }

            if (checkID in model.similar) {
                threshold = STATE_SIMILAR;
                finalCol = initialCol;
            }

            if (checkID in model.identical) {
                threshold = STATE_IDENTICAL;
                finalCol = Math.floor(NUM_COLUMNS / 2);

                /*
                 * If all members of an identical set belong to the same
                 * list, merge them; the middle column is reserved for
                 * items that are identical among both lists.
                 */
                var identical = model.getIdentical(checkID);

                for (var j = 0; j < identical.length; j++) {
                    if (model.items[identical[j]].listID !==
                                checkItem.listID) {
                        break;
                    }
                }

                if (j === identical.length) {
                    finalCol = initialCol;

                    // TODO: hack, should really support STATE_IDENTICAL

                    threshold = STATE_SIMILAR;
                }
            }
            positions[id] = {
                threshold : threshold,
                initialCol : initialCol,
                finalCol : finalCol
            };
        }
    }

    function calculatePositions() {
        // clear rows
        for (var i = 0; i < viewData.length; i++) {
            var position = positions[viewData[i]];

            if (position.hasOwnProperty("initialRow")) {
                delete position.initialRow;
            }

            if (position.hasOwnProperty("thresholdRow")) {
                delete position.thresholdRow;
            }

            if (position.hasOwnProperty("finalRow")) {
                delete position.finalRow;
            }
        }

        // (re-)calculate
        calculateRows();
        calculateKeyframes();
        
        /*
         * Shadow originals even when hidden; shadows should appear to
         * emerge from, and disappear into, their shadowed items.
         */
        for (var id in model.shadows) {
            if (id in model.hidden) {
                $.extend(true, positions[id],
                        positions[model.getShadowed(id)]);
            }
        }
    }

    function calculateRows() {
        var initialRows = [0, 0];
        var thresholdRows = [], finalRows = [];
        var bumped = [];

        for (var i = 0; i < NUM_COLUMNS; i++) {
            thresholdRows.push(0);
            finalRows.push(0);
        }

        // reset groups
        groups = [];
        var currentGroup = "";
        var previousGroup = "";

        for (var i = 0; i < viewData.length; i++) {
            var id = viewData[i];
            var position = positions[id];

            // impose group restriction on spatial position
            if (model.groupBy) {
                previousGroup = currentGroup;
                currentGroup = model.items[id].attributes[model.groupBy]
                        [model.items[id].groupByOffset];

                if (currentGroup.toLowerCase() !==
                        previousGroup.toLowerCase()) {
                    var row = Math.max.apply(null, initialRows);
                    var thresholdRow = Math.max.apply(null,
                            thresholdRows);
                    var finalRow = Math.max.apply(null, finalRows);

                    initialRows[0] = row;
                    initialRows[1] = row;

                    for (var j = 0; j < NUM_COLUMNS; j++) {
                        thresholdRows[j] = thresholdRow;
                        finalRows[j] = finalRow;
                    }

                    if (previousGroup.length > 0) {
                        var previous = groups[groups.length - 1];

                        previous.initialRowEnd = row;
                        previous.thresholdRowEnd = thresholdRow;
                    }
                    groups.push({
                        label : currentGroup,
                        initialRowStart : row,
                        thresholdRowStart : thresholdRow,
                        members : [id]
                    });
                } else {
                    groups[groups.length - 1].members.push(id);
                }
            }
            position.initialRow =
                    (model.items[id].listID === model.list1.id) ?
                    initialRows[0]++ : initialRows[1]++;
            calculateRow(id, "thresholdRow", thresholdRows, currentGroup);
            calculateRow(id, "finalRow", finalRows, currentGroup, bumped);
        }

        // re-align bumped-down items
        var height = Math.max.apply(null, finalRows);

        for (var i = 0; i < NUM_COLUMNS; i++) {
            finalRows[i] = height;
        }

        for (var i = 0; i < bumped.length; i++) {
            calculateRow(bumped[i], "finalRow", finalRows, currentGroup);
        }

        if (model.groupBy && groups.length > 0) {
            var group = groups[groups.length - 1];

            group.initialRowEnd = Math.max.apply(null, initialRows);
            group.thresholdRowEnd = Math.max.apply(null, thresholdRows);
            group.finalRowEnd = height;
        }
    }

    function calculateRow(id, type, rows, group, bumped) {
        var position = positions[id];
        var identical = model.getIdentical(id, true);
        var similar = model.getSimilar(id, true);

        // row already calculated
        if (position.hasOwnProperty(type)) {
            return;
        }

        // bump similar items down
        if (bumped) {
            if (id in model.similar) {
                bumped.push(id);
                
                return;
            } else if (id in model.identical) {
                for (var i = 0; i < identical.length; i++) {
                    if (identical[i] in similar) {
                        bumped.push(id);
                        
                        return;
                    }
                }
            }
        }

        // position self
        position[type] = rows[position.finalCol];

        // position identical items
        for (i = 0; i < identical.length; i++) {
            var relatedPosition = positions[identical[i]];

            if (!relatedPosition.hasOwnProperty(type)) {
                relatedPosition[type] = position[type];
            }
        }

        // position similar items
        rows[position.finalCol]++;

        for (var i = 0; i < similar.length; i++) {
            relatedPosition = positions[similar[i]];

            if (!relatedPosition.hasOwnProperty(type)) {
                if (model.groupBy) {
                    if (model.items[similar[i]]
                                .attributes[model.groupBy][0] === group) {
                        relatedPosition[type] =
                                rows[relatedPosition.finalCol]++;
                    }
                } else {
                    relatedPosition[type] =
                            rows[relatedPosition.finalCol]++;
                }
            }
        }

        // re-align
        if (type === "thresholdRow") {
            var height = Math.max.apply(null, rows);

            for (var i = 0; i < NUM_COLUMNS; i++) {
                rows[i] = height;
            }
        } else {
            var height = Math.max(rows[LIST_1_INDEX],
                    rows[LIST_IDENTICAL_INDEX], rows[LIST_2_INDEX]);

            for (var i = LIST_1_INDEX; i <= LIST_2_INDEX; i++) {
                rows[i] = height;
            }
        }
    }

    function calculateKeyframes() {
        keyframes = {};

        for (var i = 0; i < NUM_STATES; i++) {
            keyframes[i] = [];
            NUM_ROWS[i] = 0;

            for (var j = 0; j < NUM_COLUMNS; j++) {
                keyframes[i][j] = [];
            }
        }

        for (var i = 0; i < viewData.length; i++) {
            var id = viewData[i];
            var position = positions[id];
            var col;

            // initial position
            for (var j = 0; j < position.threshold; j++) {
                position[j] = {
                    row : position.initialRow,
                    col : position.initialCol
                };
                col = keyframes[j][position.initialCol];
                
                if (col[position.initialRow] === undefined) {
                    col[position.initialRow] = [];
                }
                col[position.initialRow].push(id);
                NUM_ROWS[j] = Math.max(NUM_ROWS[j],
                        position.initialRow + 1);
            }

            // threshold position
            for ( ; j < STATE_COMPACT; j++) {
                position[j] = {
                    row : position.thresholdRow,
                    col : position.finalCol
                };
                col = keyframes[j][position.finalCol];
                
                if (col[position.thresholdRow] === undefined) {
                    col[position.thresholdRow] = [];
                }
                col[position.thresholdRow].push(id);
                NUM_ROWS[j] = Math.max(NUM_ROWS[j],
                        position.thresholdRow + 1);
            }

            // final (compact) position
            position[j] = {
                row : position.finalRow,
                col : position.finalCol
            };
            col = keyframes[j][position.finalCol];
            
            if (col[position.finalRow] === undefined) {
                col[position.finalRow] = [];
            }
            col[position.finalRow].push(id);
            NUM_ROWS[j] = Math.max(NUM_ROWS[j], position.finalRow + 1);
        }
        
        // set empty cells to [] rather than undefined
        var height = Math.max(NUM_ROWS[LIST_1_UNIQUE_INDEX],
                NUM_ROWS[LIST_1_INDEX],
                NUM_ROWS[LIST_IDENTICAL_INDEX],
                NUM_ROWS[LIST_2_INDEX],
                NUM_ROWS[LIST_2_UNIQUE_INDEX]);
        
        for (var i = 0; i < NUM_STATES; i++) {
            for (var j = 0; j < NUM_COLUMNS; j++) {
                var col = keyframes[i][j];
                
                for (var k = 0; k < height; k++) {
                    if (col[k] === undefined) {
                        col[k] = [];
                    }
                } 
            }
        }
    } // end of calculateKeyframes
    
    /*
     * Return the number of remaining drugs to act upon
     * This varies depending on the current state
     * e.g. during Identical, drugs are merged so there seem to be fewer
     */
    function calculateNumRemaining() {
    	// return the number of page elements that are undecided
    	return $(".undecided:not(.preview):not(.shadow):visible").length;
    }
    
    // called to update the review button
    function updateReviewButton() {
    	var $numRemaining = calculateNumRemaining();
    	if($numRemaining > 0) {
    		// TODO rename
    		$(".review_button .review").html(
    									'<div style="text-align:right"><strong>sign<br/>off</strong></div>' +
    									'<span class="num_remaining">' + 
    									$numRemaining + '</span><br/>left<br/>' +
    									'<div class="patient_name">' + model.patientLastName 
    									+ ', ' + model.patientFirstName 
    									+ ' (' + model.patientAge 
    									+ model.patientGender + ')</div>');
			$(".review_button .review").addClass("disabled_button");
    	} else {
    		$(".review_button .review").html('<div style="text-align:right"><strong>sign<br/>off</strong></div>' +
    									'<div class="patient_name">' + model.patientLastName 
    									+ ', ' + model.patientFirstName 
    									+ ' (' + model.patientAge 
    									+ model.patientGender + ')</div>');
			$(".review_button .review").removeClass("disabled_button");   		
    	}
    }
        
        
    // animation /////////////////////////////////////////////////////////
    function redraw(sort, filter, immediate) {
        /*
         * Tweak proportions.
         * 
         * Set position relative to the bottom primarily so that content
         * grows / shrinks from top edge, making room for the options
         * panel without occluding anything.
         * 
         * Debatable whether the main app should have to crunch for the
         * options panel: overlapping might make it harder to forget that
         * it's open (and occupying unnecessary space).
         */
        var headerHeight = $(".banner").outerHeight(true);
        var optionsPanelHeight = $(".options").hasClass("active") ?
                $(".options-panel").outerHeight(true) : 0;
        var contentMarginTop =
                parseFloat($("#app > .content").css("margin-top"));
        var contentMarginBottom =
                parseFloat($("#app > .content").css("margin-bottom"));
        var detailHeight = $("#app > .detail").outerHeight(true);
        var brandingHeight = $(".branding").outerHeight(true);
        var footerHeight = detailHeight + brandingHeight;
        var tabHeight = $(".add").outerHeight(true);
        
        $("#app > .content").css("bottom", footerHeight);
        $("#app > .content").height($(window).height() - headerHeight -
                optionsPanelHeight - footerHeight - contentMarginTop -
                contentMarginBottom);
        
        $(".add").css("top", headerHeight + contentMarginTop +
                topOffset() + optionsPanelHeight + tabHeight * 0.5);
        $(".edit").css("top", headerHeight + contentMarginTop +
                topOffset() + optionsPanelHeight + tabHeight * 1.75);
        
        // TODO: move scrolltip adjustment?

        adjustScrolltips();

        if (immediate) {
            viewData = model.viewData(sort, filter);
            calculatePositions();

            $("#app .item").each(function() {
                animateItem(this.id, state,
                        visible.animationDelay / 4);
            });
        } else {
            viewData = model.viewData(sort, filter);
            calculatePositions();
            changeState(state);
        }
    }
    
    function changeState(toState) {
        if (state === toState || version === VERSION_BASELINE) {
            transition(state, state);
        } else {
            var transitionDuration = visible.animationDuration;
            
            if (model.groupBy && toState > STATE_SIMILAR) {
                // grouping doesn't make sense in STATE_COMPACT
                toState = STATE_SIMILAR;
            }
            
            if (visible.animationSpeed === ANIMATION_SPEED_0) {
                // no animation, jump to destination state
                transition(toState, toState);
            } else {
                var offset = (state < toState) ? 1 : -1;
                var i = state;
                
                while (i !== toState) {
                    transition(state, i + offset);
                    i += offset;
                }
                transitionDuration += transitionDelay(state,
                        toState - offset);
            }
            logger.log(logger.EVENT_STATE_CHANGE, 
            		"start:" + stateIndexToName(state) +
            		",end:" + stateIndexToName(toState) +
                    ",duration(ms):" + transitionDuration);
            state = toState;
        }
    }
    
    function transition(from, to) {
        var delay = 0;

        if (from < to) {
            delay = transitionDelay(state, to - 1);
        } else if (from > to) {
            delay = transitionDelay(state, to + 1);
        }
        
        // TODO: track timers to allow queue canceling
        
        setTimeout(function() {
            adjustAnimationControls(to);
            adjustBackdrop(to);
            adjustCellDimensions();
            adjustColumnHeaders(from, to);
            adjustGroupLabels(from, to);
            adjustDifferenceHighlights(to);

            // adjust row striping
            $(".backdrop td").css("background", "");
            
            var $tbody = $(".backdrop tbody");
            var height = $tbody.children().length;

            for (var i = 0; i < height; i++) {
                $tbody.children(":nth-child(" + (i + 1) + ")")
                        .css("background", "");
            }

            if (model.groupBy) {
                var color = WHITE;

                for (var i = 0; i < groups.length; i++) {
                    var group = groups[i];
                    var start = "thresholdRowStart";
                    var end = "thresholdRowEnd";
                    var threshold = STATE_SIMILAR;

                    for (var j = 0; j < group.members.length; j++) {
                        var id = group.members[j];

                        if (positions[id].threshold < threshold) {
                            threshold = positions[id].threshold;
                        }
                    }

                    if (state < threshold) {
                        start = "initialRowStart";
                        end = "initialRowEnd";
                    }

                    if (group.hasOwnProperty(start) &&
                            group.hasOwnProperty(end)) {
                        for (var j = group[start]; j < group[end]; j++) {
                            $tbody.children(":nth-child(" + (j + 1) + ")")
                                    .css("background", color);
                        }
                        color = (color === WHITE) ? NEUTRAL_GRAY : WHITE;
                    }
                }
            } else if (to === STATE_COMPACT) {
                // paint white first
                $(".backdrop td").css("background", WHITE);
                
                // block unique and identical
                var height = Math.max(model.unique1.length,
                        model.unique2.length,
                        keyframes[to][LIST_IDENTICAL_INDEX].length);
                var block = [LIST_1_UNIQUE_INDEX, LIST_2_UNIQUE_INDEX,
                        LIST_IDENTICAL_INDEX];
                
                for (var i = 0; i < height; i++) {
                    for (var j = 0; j < block.length; j++) {
                        $tbody.children(":nth-child(" + (i + 1) + ")")
                                .children(":nth-child(" +
                                (block[j] + 1) + ")")
                                .css("background", NEUTRAL_GRAY);
                    }
                }
                
                // stripe similar
                var color = CONTRAST_GRAY;
                var col = keyframes[to][LIST_1_INDEX];
                var striped = {};
                
                for (var i = 0; i < col.length; i++) {
                    var ids = col[i];
                    
                    for (var j = 0; j < ids.length; j++) {
                        var id = ids[j];
                        
                        if (id !== undefined &&
                                striped[id] === undefined) {
                            // color single item
                            var pos = positions[id][to];
                            
                            $tbody.children(":nth-child(" +
                                    (pos.row + 1) + ")").children("td")
                                    .css("background", color);
                            striped[id] = true;
                            
                            // color related items accordingly
                            var related = model.getRelated(id);
                            
                            for (var j = 0; j < related.length; j++) {
                                id = related[j];
                                pos = positions[id][to];
                                
                                $tbody.children(":nth-child(" +
                                        (pos.row + 1) + ")")
                                        .children("td")
                                        .css("background", color);
                                striped[id] = true;
                            }
                            color = color === WHITE ?
                                    CONTRAST_GRAY : WHITE;
                        }
                    }
                }
            } else {
                // stripe rows for legibility
                $tbody.children(":nth-child(odd)").css("background",
                        WHITE);
                $tbody.children(":nth-child(even)").css("background",
                        NEUTRAL_GRAY);
            }

            // animate items
            if (to > from) {
                if (to === STATE_IDENTICAL) {
                    animateIdentical(to);
                } else if (to === STATE_UNIQUE) {
                    animateUnique(to);
                } else {
                    animateDefault(to);
                }
            } else {
                animateDefault(to);
            }
        }, delay);

    }
    
    function animateIdentical(toState) {
        var i = 0;
        var checked = {}, animated = {};

        for (var id in model.identical) {
            if (checked[id] === undefined) {
                var set = model.identical[id];
                
                for (var j = 0; j < set.length; j++) {
                    var checkID = set[j];
                    
                    if (model.items[checkID].isShadowed) {
                        set = set.concat(model.getShadows(checkID));
                    }
                }
                checked[id] = true;

                for (var j = 0; j < set.length; j++) {
                    checked[set[j]] = true;
                }
                animateSet(set, toState, (i > 0) ?
                        (i * visible.animationDuration) +
                        (i * visible.animationDelay) : 0, animated);
                i++;
            }
        }
    }
    
    function animateUnique(toState) {
        var unique1 = model.unique1, unique2 = model.unique2;
        
        for (var i = 0; i < model.unique1.length; i++) {
            var id = model.unique1[i];
            var item = model.items[id];
            
            if (item.isShadowed) {
                unique1 = unique1.concat(model.getShadows(id));
            }
        }
        
        for (var i = 0; i < model.unique2.length; i++) {
            var id = model.unique2[i];
            var item = model.items[id];
            
            if (item.isShadowed) {
                unique2 = unique2.concat(model.getShadows(id));
            }
        }
        
        // animate left first, then right
        animateSet(unique1, toState, 0)
        animateSet(unique2, toState, visible.animationDuration +
                visible.animationDelay);
    }
    
    function animateDefault(toState) {
        $("#app .item").each(function() {
            animateItem(this.id, toState);
        });
    }
    
    function animateSet(set, toState, delay, animated) {
        setTimeout(function() {
            for (var i = 0; i < set.length; i++) {
                var id = set[i];

                if (animated) {
                    if (animated[id] === undefined) {
                        animated[id] = true;
                        animateItem(id, toState);
                    }
                } else {
                    animateItem(id, toState);
                }
            }
        }, delay);
    }
    
    function animateItem(id, toState, duration) {
        var $item = $("#" + id);
        var position = (toState !== undefined) ?
                offsetToPosition(positions[id][toState]) :
                offsetToPosition(positions[id]);
        
        // TODO: animate() or transition()?
        
        $item.transition({
            left : position.x,
            top : position.y,
            duration: duration ? duration : visible.animationDuration,
            easing: CUSTOM_EASE_IN
        });
        
        // handle identical overlap
        if (id in model.identical) {
            if (state < STATE_IDENTICAL) {
                if (id in model.hidden) {
                    delete model.hidden[id];
                    visible.toggleItem($item, visible.toggleOnDelay,
                            true);
                }
            } else {
                if (model.identical[id].indexOf(parseFloat(id)) !== 0) {
                    model.hidden[id] = true;
                    visible.toggleItem($item, visible.toggleOffDelay,
                            false);
                }
            }
        }
    }
    
    function topOffset() {
        return $(".backdrop th").outerHeight(true);
    }

    function columnWidth() {
        return $(".backdrop td").outerWidth(true);
    }

    function rowHeight() {
        // include padding but exclude margin
        return $(".backdrop td").outerHeight();
    }
    
	// given offset information (which row/column), convert into (x,y) position info
    function offsetToPosition(offset) {
    	var twoColOffset = 0; // offset for 2 column versions to make lists closer together
    	if(version != VERSION_FULL) {
    		twoColOffset = $(".backdrop thead").outerWidth() * 0.1;
    		if(offset.col == 3)
    			twoColOffset = twoColOffset * -1;
    	}
        return {
            x : offset.col * columnWidth() + twoColOffset,
            y : (offset.row * rowHeight()) + topOffset()
        };
    }

    function transitionDelay(from, to) {
        var delay = 0;

        if (from < to) {
            for (var i = from + 1; i <= to; i++) {
                if (i === STATE_IDENTICAL) {
                    delay += transitionIdenticalDelay();
                } else if (i === STATE_UNIQUE) {
                    delay += 2 * (visible.animationDuration +
                            visible.animationDelay);
                } else {
                    delay += visible.animationDuration +
                            visible.animationDelay;
                }
            }
        } else if (from > to) {
            for (var i = from; i >= to + 1; i--) {
                delay += visible.animationDuration +
                        visible.animationDelay;
            }
        }
        return delay;
    }
    
    function transitionIdenticalDelay() {
        var numSets = 0;
        var checked = {};

        for (var id in model.identical) {
            if (checked[id] === undefined) {
                var set = model.getIdentical(id);
                checked[id] = true;

                for (var j = 0; j < set.length; j++) {
                    checked[set[j]] = true;
                }
                numSets++;
            }
        }
        return (numSets * visible.animationDuration) +
                (numSets * visible.animationDelay);
    }
    
    function setAnimationSpeed(speed) {
        // adjust speed
        visible.animationSpeed = speed;
        $("select[name='speed']").val(speed);
        
        if (speed === ANIMATION_SPEED_0) {
            // use neutral (median) animation speed for single transition
            visible.animationDuration = BASE_ANIMATION_DURATION *
                    ANIMATION_SPEED_COEFFICIENTS[
                    Math.floor(ANIMATION_SPEED_COEFFICIENTS.length / 2)];
        } else {
            var coefficient = speed.replace(/[^0-5]+/g, "");
            visible.animationDuration = BASE_ANIMATION_DURATION *
                    ANIMATION_SPEED_COEFFICIENTS[coefficient];
        }
        
        // adjust delays accordingly
        visible.animationDelay = visible.animationDuration * 0.75;
        visible.toggleOnDelay = visible.animationDelay / 4;
        visible.toggleOffDelay = visible.animationDelay * 1.5;
        
        adjustAnimationControls(state);
        
        logger.log(logger.EVENT_ANIMATION_SPEED_CHANGE, visible.animationSpeed);
    }
    
    
    // interface adjustment //////////////////////////////////////////////
    function adjustBackdrop(toState) {
        // redraw everything: avoid odd white stripe when end row changes
        $(".backdrop tbody tr").remove();

        for (var i = 0; i < NUM_ROWS[toState]; i++) {
            var $tr = $("<tr></tr>");

            for (var j = 0; j < NUM_COLUMNS; j++) {
                $tr.append("<td></td>");
            }
            $(".backdrop tbody").append($tr);
        }
    }
    
    function adjustColumnHeaders(from, to) {
        toggleHeader(LIST_1_UNIQUE_INDEX, to < STATE_UNIQUE ? false :
                true);
        toggleHeader(LIST_IDENTICAL_INDEX, to < STATE_IDENTICAL ? false :
                true);

        // match item animation: list 1 unique, then list 2 unique
        setTimeout(function() {
            toggleHeader(LIST_2_UNIQUE_INDEX, to < STATE_UNIQUE ? false :
                    true);
        }, from >= to ? 0 :
                visible.animationDuration + visible.animationDelay);

        toggleConditionalHeaders(to < STATE_SIMILAR ? false : true);
    }
    
    /*
     * Given a state, adjust animation controls appropriately
     * "state" is one of the defined constants above
     */
    function adjustAnimationControls(state) {
        $(".compare").removeClass("inactive");
        $(".separate, .identical, .unique, .similar, .compact")
                .removeClass("active inactive");
        var active = {};
        var inactive = {};
        
        // inactive
        if (visible.animationSpeed === ANIMATION_SPEED_0) {
            inactive[STATE_IDENTICAL] = true;
            inactive[STATE_UNIQUE] = true;
            inactive[STATE_SIMILAR] = true;
        }
        
        if (model.groupBy) {
            inactive[STATE_COMPACT] = true;
            delete inactive[STATE_SIMILAR];
        }
        
        if (version === VERSION_BASELINE ||
                version === VERSION_LINK_ONLY) {
            for (var i = STATE_IDENTICAL; i < NUM_STATES; i++) {
                inactive[i] = true;
            }
            $(".compare").addClass("inactive");
        }
        
        for (var stateIndex in inactive) {
            $("." + stateIndexToName(parseInt(stateIndex)))
                    .addClass("inactive");
        }
        $("." + stateIndexToName(state)).removeClass("ianctive")
                .addClass("active");
    }
    
    function adjustGroupLabels(from, to) {
        $(".groups .label").remove();
        
        if (model.groupBy) {
            for (var i = 0; i < groups.length; i++) {
                var group = groups[i];
                var row = group.thresholdRowStart;
                var threshold = STATE_SIMILAR;
                var $label = $("<span></span>");
                
                for (var j = 0; j < group.members.length; j++) {
                    var id = group.members[j];
                    
                    if (positions[id].threshold < threshold) {
                        threshold = positions[id].threshold;
                    }
                }
                
                if (state < threshold) {
                    row = group.initialRowStart;
                }
                $label.attr("class", "label");
                $label.text(group.label);
                $label.css("left", 0);
                $label.css("top", topOffset() + (row * rowHeight()));
                $(".groups").append($label);
            }
            $(".groups .label").addClass("show");
        }
    }
    
    function adjustCellDimensions() {
        // adjust padding
        $("#app .item").css("padding-top",
                model.groupBy ? "1em" : "0.6em");
        $("#app .item").css("padding-bottom",
                model.groupBy ? 0 : "0.4em");
        
        // adjust height
        if (displayDetails) {
            $("#app .item .detail").fadeIn();
            $("#app .item, #review td").height($.browser.mozilla ? "3em" :
                    "2.5em");
        } else {
            $("#app .item .detail").fadeOut();
            $("#app .item, #review td").height($.browser.mozilla ?
                    "1.75em" : "1.25em");
        }
        $(".backdrop td").height($.browser.mozilla ?
                $("#app .item").outerHeight() : $("#app .item").height());
    }
    
    // given a state, adjust difference highlights on items
    function adjustDifferenceHighlights(toState) {
        if (toState < STATE_SIMILAR) {
            $(".difference").removeClass("highlight");
        } else {
            if (model.displayName !== model.RECORDED_NAME) {
                $(".name.difference").removeClass("highlight");
                $(".undecided .difference:not(.name)")
                        .addClass("highlight");
            } else {
                $(".undecided .difference").addClass("highlight");
            }
        }
    }
    
    function adjustScrolltips() {
        var bottom = $("#app > .detail").outerHeight(true) +
                $(".branding").outerHeight(true) +
                parseFloat($("#app > .content").css("margin-bottom"));
        
        // hand-tweaked positions, adjust as appropriate
        $(".up").css("top", topOffset());
        $(".down").css("bottom", bottom);
    }
    
    function toggleHeader(index, show) {
        var $header = $(".backdrop th:nth-child(" + (index + 1) +
                ") .col-header");
        
        if (show) {
            $header.fadeIn(visible.toggleOnDelay);
        } else {
            $header.fadeOut(visible.toggleOnDelay);
        }
    }
    
    function toggleConditionalHeaders(show) {
        if (show) {
            $(".backdrop th:nth-child(" +
                    (LIST_1_INDEX + 1) + ") .name .conditional")
                    .fadeIn(visible.toggleOnDelay);
            $(".backdrop th:nth-child(" + (LIST_2_INDEX + 1) +
                    ") .name .conditional").fadeIn(visible.toggleOnDelay);
        } else {
            $(".backdrop th:nth-child(" +
                    (LIST_1_INDEX + 1) + ") .name .conditional")
                    .fadeOut(visible.toggleOffDelay);
            $(".backdrop th:nth-child(" + (LIST_2_INDEX + 1) +
                    ") .name .conditional")
                    .fadeOut(visible.toggleOffDelay);
        }
    }
    
    function stateIndexToName(index) {
        switch (index) {
            case STATE_SEPARATE:
                return "separate";
            case STATE_IDENTICAL:
                return "identical";
            case STATE_UNIQUE:
                return "unique";
            case STATE_SIMILAR:
                return "similar";
            case STATE_COMPACT:
                return "compact";
        }
        return undefined;
    }
    
    
    // item events ///////////////////////////////////////////////////////
    function mousedownHandler(event) {
    	logger.log(logger.EVENT_CLICKED, "item");
    	
        var dst;

        // TODO: very useful for debugging, keep until final ship
        console.log(this.id);
        visible.editID = this.id;

        if (event.which === 1) {                // left click
            if ($(this).hasClass("undecided")) {
                dst = "accepted";
            } else if ($(this).hasClass("accepted")) {
                dst = "rejected";
            } else if ($(this).hasClass("rejected")) {
                dst = "undecided";
            }
        } else if (event.which === 3) {         // right click
            if ($(this).hasClass("undecided")) {
                dst = "rejected";
            } else if ($(this).hasClass("accepted")) {
                dst = "undecided";
            } else if ($(this).hasClass("rejected")) {
                dst = "accepted";
            }
        }

        if (dst) {
            processItem(this.id, dst, version === VERSION_BASELINE);
            
            if (model.afterAction === model.AFTER_ACTION_REMOVE) {
                redraw(true, true);
            }
        }
        
        // some item was changed, update sign off button's undecided count
        updateReviewButton();
    }

    function mouseoverHandler() {
        var id = this.id;
        var item = model.items[id];
        
        // hover appropriate item(s)
        if (version === VERSION_FULL || version === VERSION_LINK_ONLY) {
            var toHover = model.getRelatedSet(id, model.multigroup);
            
            for (var i = 0; i < toHover.length; i++) {
                var hoverID = toHover[i];
                hoverSet[hoverID] = true;
                hoverItem(hoverID, true);
            }
        } else {
            hoverSet[id] = true;
            hoverItem(id, true);
        }
        hoverScrolltips(id, true);

        // update item detail
        var $detail = $(".detail .content");
        var fullDescription = item.name + " | ";

        for (var attributeName in model.attributes) {
            if (attributeName in item.attributes) {
                var values = item.attributes[attributeName];
                var valuesString = "";

                if (values.length > 1) {
                    values = item.attributes[attributeName];

                    for (var i = 0; i < values.length; i++) {

                        if (attributeName === model.ATTR_SUBITEM) {
                            var subitem = values[i];
                            var dose = subitem
                                    .attributes[model.ATTR_DOSE];
                            
                            valuesString += subitem.name + (dose ?
                                    " " +
                                    subitem.attributes[model.ATTR_DOSE] :
                                    "");
                        } else {
                            valuesString += values[i];
                        }
                        valuesString += ", ";
                    }
                    valuesString = valuesString.slice(0, -2);
                } else {
                    valuesString = values.toString();
                }
                
                if (valuesString.trim() !== "") {
                    fullDescription += valuesString + " | ";
                }
            }
        }
        fullDescription = fullDescription.slice(0, -3);
        var editedExplanation = "";
        if($(this).hasClass("modified")) {
        	editedExplanation = " | (Note: * = modified item)"
        }
        $detail.text(fullDescription + editedExplanation);
    }

    function mouseoutHandler() {
        for (var id in hoverSet) {
            hoverItem(id, false);
        }
        hoverSet = {};
        hoverScrolltips(this.id, false);
        
        $(".detail .content").text("Nothing to display.");
    }
    
    // given an id and whether this is on mouseover, hover the item
    function hoverItem(id, mouseover) {
        var item = model.items[id];
        var $item = $("#" + id);
        
        if (mouseover) {
            $item.addClass("item-hover");
            
            if(item.isShadow)
            	$item.css("opacity", "0.5");
        	else
            	$item.css("opacity", "1"); // TODO verify

            if ($item.hasClass("undecided")) {
                $item.addClass("undecided-hover");
            }
            
            // for 2 column w/ links, highlight differences on mouseover
            if(version === VERSION_LINK_ONLY)
            	$("#" + id + " .difference").addClass("highlight");
            
        } else {
            $item.removeClass("item-hover");

            if ($item.hasClass("undecided")) {
                $item.removeClass("undecided-hover");
            }
            
            // for 2 column w/ links, highlight differences on mouseover
            if(version === VERSION_LINK_ONLY)
            	$("#" + id + " .difference").removeClass("highlight");
        }
    }

    function hoverScrolltips(id, mouseover) {
        $(".scrolltip").removeClass("show");

        if (mouseover) {
            var checkItems = (version === VERSION_FULL ||
                    version === VERSION_LINK_ONLY) ?
                    model.getRelatedSet(id, model.multigroup) :
                    model.getShadowSet(id);
            var numObscured = {
                "neither" : 0,
                "above" : 0,
                "below" : 0
            };
            
            for (var i = 0; i < checkItems.length; i++) {
                numObscured[isOffscreen(checkItems[i])]++;
            }
            
            if (numObscured["above"] > 0) {
                $(".up").text("...more (" + numObscured["above"] + ")");
                $(".up").addClass("show");
            }
            
            if (numObscured["below"] > 0) {
                $(".down").text("...more (" + numObscured["below"] + ")");
                $(".down").addClass("show");
            }
        }
    }
    
    function isOffscreen(id) {
        var status = "neither";
        var $content = $(".content");
        var divHeight = $(".item").outerHeight();
        var topEdge = $content.scrollTop();
        var bottomEdge = topEdge + $content.outerHeight();
        var position = offsetToPosition(positions[id][state]);
            
        if ((position.y + divHeight) < topEdge ||
                topEdge > (position.y + divHeight * 0.75)) {
            status = "above";
        } else if (position.y > bottomEdge ||
                    (position.y + divHeight * 0.25) > bottomEdge) {
            status = "below";
        }
        return status;
    }

    
    // item action ///////////////////////////////////////////////////////
    function resetDecisions() {
        model.accepted = [];
        model.rejected = [];
        model.undecided =
                model.list1.source.concat(model.list2.source);
        
        $(".item").removeClass("accepted rejected")
                .addClass("undecided");
        resetLinkActionFlags();

        state = STATE_SEPARATE;
    }
    
    function resetLinkActionFlags() {
        linkAction = {};
        
        for (var id in model.items) {
            linkAction[id] = true;
        }
    }
    
    function processItem(id, dst, nolink) {
        // don't trigger meaningless linked actions
        if (!$("#" + id).hasClass(dst)) {
            if (nolink) {
                actOnItem(id, dst);
            } else {
                // act on shadow = act on item
                actOnSet(model.items[id].isShadow ?
                        model.getShadowed(id) :
                        id, dst);
            }
        }
    }
    
    // called to handle column actions (like "keep", "reject", "clear")
    function processColumn(event) {
    	logger.log(logger.EVENT_CLICKED, "column_action");
    	logger.log(logger.EVENT_COLUMN_ACTION, "dst:" + event.data.dst 
											+ ",state:" + state
											+ ",column:" + event.data.index);
    	
        var col = keyframes[state][event.data.index];
		console.log(state + " " + event.data.index);
		
		
		
        for (var i = 0; i < col.length; i++) {
            var ids = col[i];
            
            for (var j = 0; j < ids.length; j++) {
                // only act on items in this particular column
                //  and only act on undecided ones
                if(event.data.dst == 'undecided' 
                	|| $("#" + ids[j]).hasClass('undecided')) {
                	processItem(ids[j], event.data.dst, true);		
            	} 
                
            }
        }
        
        if (col.length > 0 &&
                    model.afterAction === model.AFTER_ACTION_REMOVE) {
            redraw(true, true);
        }
        
        // update review button's number of undecided
        updateReviewButton();
    }
    
    function actOnSet(id, dst) {
        var identical = model.getIdentical(id);
        var similar = model.getSimilar(id);
        
        // always link actions on identical sets
        for (var i = 0; i < identical.length; i++) {
            actOnItem(identical[i], dst === "accepted" ? "rejected" :
                    dst);
        }
        
        // only link actions on similar sets for the first click
        if (linkAction[id] && similar.length > 0) {
            for (var i = 0; i < similar.length; i++) {
                actOnItem(similar[i], dst === "undecided" ? "undecided" :
                        "rejected");
            }
            
            if (dst === "rejected") {
                var related = model.getRelated(id);
                var next = related[0];
                
                if (next in model.identical) {
                    // ensure that next is always visible
                    next = model.identical[next][0];
                }
                actOnItem(next, "accepted");
            }
        }
        actOnItem(id, dst);
    }
    
    function actOnItem(id, dst) {
        // act on item
        decide(id, dst);
        
        // act on shadows
        var shadows = model.getShadows(id);
        
        for (var i = 0; i < shadows.length; i++) {
            decide(shadows[i], dst);
        }
    }
    
    function decide(id, dst) {
        var $item = $("#" + id);
        var $differences = $("#" + id + " .difference");
        var checkID = model.items[id].isShadow ? model.getShadowed(id) :
                id;
        
        // update model
        var src = "undecided";
        
        if ($item.hasClass("accepted")) {
            src = "accepted";
        } else if ($item.hasClass("rejected")) {
            src = "rejected";
        }
        model.decide(id, src, dst);
        
        // update view
        $item.removeClass("accepted rejected undecided undecided-hover");
        
        if (dst === "undecided") {
            if (id in hoverSet) {
                $item.addClass("undecided-hover");
            }
            
            if (state >= STATE_SIMILAR) {
                $differences.addClass("highlight");
            }
        } else {
            $differences.removeClass("highlight");
        }
        $item.addClass(dst);
        
        /*
         * Actions on members of identical sets are always linked.
         * 
         * Only the first action on a similar set is linked.
         */
        linkAction[id] = checkID in model.identical ? true : false;
    }
    
    function updateItem(id, allNames, attributes) {
        // update model
        var item = model.items[id];
        
        // verify was changed in some way
        var unchanged = allNames['recorded'] == item.name;
        for (var aN in attributes) {
        	var oldValue = item.attributes[aN] == undefined ? "" : item.attributes[aN].toString();
        	var newValue = attributes[aN] == undefined ? "" : attributes[aN].toString();
        	unchanged = unchanged && oldValue == newValue;
        }
        
        if(!unchanged) {
	        item.setName(allNames);
			item.isModified = true;
			
	        for (var attributeName in attributes) {
	            item.attributes[attributeName] = attributes[attributeName];
	        }
	        
	        // update view
	        $("#" + id + " .name").text(item.name);
	        $("#" + id + " .detail").remove();
	        
	        var newDetailHTML = "<div class='detail'>";
	        var differences = [];
	        
	        if (id in model.similar) {
	            differences = model.similar[item.id].differences;
	        }
	        
	        for (var attribute in item.attributes) {
	            if (model.attributes[attribute].display) {
	                newDetailHTML += "<span" +
	                        (differences.indexOf(attribute) !== -1 ?
	                        " class='difference'" : "") + ">" +
	                        item.attributes[attribute].toString() +
	                        "</span>"
	            }
	        }
	        newDetailHTML += "</div>";
	        
	        $("#" + id).addClass("modified");
	        $("#" + id).append($(newDetailHTML));
        } else {
        	item = false; // return falsey value so no updates made
        }
        return item;
    }
    
    
    // TODO: sort ////////////////////////////////////////////////////////
    function saveModifiedItem(type) {
        var allNames = {};
        var attributes = {};
        
        var name = $("#modify input[name='name']").val() + "";
        var route = $("#modify input[name='route']").val() + "";
        var frequency = $("#modify input[name='frequency']").val() + "";
        var dose = $("#modify input[name='dose']").val() + "";
        var instructions =
                $("#modify textarea[name='instructions']").val() || "";
        var drugClass = "unknown";
        
        allNames = {
            recorded: name,
            generic: name,
            brand: name
        };
        attributes[model.ATTR_DOSE] = [dose];
        attributes[model.ATTR_ROUTE] = [route];
        attributes[model.ATTR_FREQUENCY] = [frequency];
        attributes[model.ATTR_INSTRUCTIONS] = [instructions];
        
        if (type === "edit") {
            var newItem =
                    updateItem(visible.editID, allNames, attributes);
            
            // if nothing changed, newItem will be false
            if(newItem) {
            	
	            // update shadows or shadowed items, if any
	            var updateIDs = [];
	            
	            if (newItem.isShadow) {
	                updateIDs = updateIDs
	                        .concat(model.getShadowed(visible.editID));
	            } else {
	                updateIDs = updateIDs
	                        .concat(model.getShadows(visible.editID));
	            }
	            
	            for (var i = 0; i < updateIDs.length; i++) {
	                updateItem(updateIDs[i], allNames, attributes);
            	}
            }
        } else {
            // update model - needs a name at least
            if(name) { 
	            var newItem = {};
	            
	            attributes[model.ATTR_DRUG_CLASS] = [drugClass];
	            newItem = new ListItem(model.nextID++, model.list2.id,
	                    allNames, attributes);
	            newItem.groupByOffset = 0;
	            newItem.isModified = true;
	            newItem.isShadow = false;
	            newItem.isShadowed = false;
	
	            model.items[newItem.id] = newItem;
	            model.list2.source.push(newItem.id);
	            model.unique2.push(newItem.id);
	            model.accepted.push(newItem.id);
	            positions[newItem.id] = {
	                threshold: STATE_UNIQUE,
	                initialCol: LIST_2_INDEX,
	                finalCol: LIST_2_UNIQUE_INDEX
	            };
	            
	            // update view
	            var htmlItem = "<div id='" + newItem.id +
	                        "' class='item accepted modified'>" +
	                    "<div class='header'>" +
	                        "<span class='name'>" + name + "</span>" +
	                    "</div>" +
	                    "<div class='detail'>";
	            
	            for (var attribute in newItem.attributes) {
	                if (model.attributes[attribute].display) {
	                    htmlItem += "<span class='" +
	                            newItem.attributes[attribute].toString() +
	                            "'>" +
	                            newItem.attributes[attribute].toString() +
	                            "</span>";
	                }
	            }
	            htmlItem += "</div></div>";
	            $(".items").append($(htmlItem));
	            
	            // prepare actions
	            $("#" + newItem.id).mousedown(mousedownHandler)
	                    .mouseover(mouseoverHandler)
	                    .mouseout(mouseoutHandler)
	                    .bind("contextmenu", function() {
	                        // free right click for item action
	                        return false;
	                    });
            }
        }
        redraw(true, true);
    }
    
    function resetModifyPanel(type) {
    	logger.log( logger.EVENT_MODIFY_PANEL_START, "type:" + type);
        var attributes = {
            name : "",
            dose : "",
            route : "",
            frequency : "",
            instructions: ""
        };
        
        // nothing to respond to
        $("#modify .response").removeClass("show");

        // panel title
        $("#modify h1").text(type);
        
        // preview and form contents
        if (type === "edit") {
            var item = model.items[visible.editID];
            
            attributes = {
                name : item.name,
                dose : item.attributes[model.ATTR_DOSE],
                route : item.attributes[model.ATTR_ROUTE],
                frequency : item.attributes[model.ATTR_FREQUENCY],
                instructions : model.ATTR_INSTRUCTIONS in item.attributes ?
                        item.attributes[model.ATTR_INSTRUCTIONS] : ""
            };
        }
        
        for (var attribute in attributes) {
            var value = attributes[attribute].toString();
            
            if (attribute === "instructions") {
                $("textarea[name='instructions']").val(value);
            } else {
                $("input[name='" + attribute + "']").val(value);
                $(".preview ." + attribute).text(value);
            }
        }
    }
    
    
    function updatePreview($field) {
        $(".preview ." + $field.attr("name")).text($field.val());
    }
    
    function populateColumn(arr, rows) {
        for (var i = 0; i < arr.length; i++) {
            rows[i] += populateReviewItem(arr[i]);
        }
        
        for ( ; i < rows.length; i++) {
            rows[i] += "<td></td>";
        }
        return rows;
    }
    
    function populateReviewItem(id) {
        var item = model.items[id];
        
        var htmlItem = "<td class='item" +
                ($("#" + id).hasClass("modified") ? " modified" : "") + "'>" +
                "<div class='header'>" +
                    "<span class='name'>" + item.name + "</span>" +
                "</div>" +
                "<div class='detail'>";
        
        for (var attribute in item.attributes) {
            if (model.attributes[attribute].display) {
                htmlItem += "<span class='" + item.attributes[attribute]
                        .toString() + "'>" +
                        item.attributes[attribute].toString() + "</span>";
            }
        }
        return htmlItem += "</div></td>";
    }
    
    // save decisions into session storage for summary.html
    function saveItemActions() {
        var stop = "", start = "", cont = "";
        
        for (var id in model.items) {
            var item = model.items[id];
            
            if (item.isShadow) {
                continue;
            }
            var $item = $("#" + id);
            var frequency = item.attributes[model.ATTR_FREQUENCY].toString();
            
            // expand abbreviations
            frequency = frequency.replace(/w/i, " weeks");
            frequency = frequency.replace(/h/i, " hours");
            frequency = frequency.replace(/B\.?I\.?D\.?/i, "two times as day");
            frequency = frequency.replace(/T\.?I\.?D\.?/i,
                    "three times a day");
            frequency = frequency.replace(/Q\.?I\.?D\.?/i, "four times a day");
            frequency = frequency.replace(/q/i, "every ");
            
            var itemString = item.listID + "\t" + item.name + "\t" + frequency + "\t" +
                    item.attributes[model.ATTR_DOSE] + "\t" + 
                    item.attributes[model.ATTR_ROUTE] + "\n";
            
            if (item.listID === model.list1.id && $item.hasClass("accepted")) {
                cont = cont + itemString;
            } else if (item.listID === model.list1.id &&
                    $item.hasClass("rejected")) {
                stop = stop + itemString;
            } else if (item.listID === model.list2.id &&
                    $item.hasClass("accepted")) {
                start = start + itemString;
            }
        }
        stop = stop.trimRight();
        start = start.trimRight();
        cont = cont.trimRight();
        
        utils.setStorageItem("stop", stop);
        utils.setStorageItem("start", start);
        utils.setStorageItem("continue", cont);
        
        // save patient name
        utils.setStorageItem("patient_name", model.patientLastName + ", " + model.patientFirstName);
    }
    
    function getVersionShortName(version) {
		switch(version) {
			case VERSION_FULL: return "FULL";
    		case VERSION_BASELINE: return "BASE";
    		case VERSION_LINK_ONLY: return "LINK";
		}
		return undefined;
	};
    
    
    
    // updates cycling/options and does redraw
    function updateGroupBy(toGroup) {
    	
    	var $curr_groupby = model.groupBy;
    	
    	// update cycling if toGroup is valid
    	var $next_val = "";
    	var $next_name = "";
    	
    	// if going to group by drug class, next one is none
    	if(toGroup == "__ATTR_DRUG_CLASS__") {
    		$next_val = "";
    		$next_name = "none";
    	} else {	// otherwise next is drug class
    		$next_val = "__ATTR_DRUG_CLASS__";
    		$next_name = "drug class";
    	}
    	
		// update the grouping in the options (using value)
		$("select[name='groupBy']").val(toGroup);
		
		// update the cycle group button's text
		$(".cycle_groupby").html("group by: " + $next_name); 
		$(".cycle_groupby").attr("value", $next_val);
		
		// update mpdel
        model.groupBy = toGroup;

		// force; grouping in compact looks strange
        if (state > STATE_SIMILAR) {
            state = STATE_SIMILAR;
        }
        
        redraw(true, true);
    }
    
    // displays the import dialog for physicians to enter end state information
    function openImportDialog() {
		$("#import-data").val(""); // reset contents of text field
    	$("#import-dialog").dialog("open");
    }
    
    // expose interface //////////////////////////////////////////////////
    return visible;
}(window.controller = window.controller || {}, $, undefined);
